@@ -244,105 +244,146 @@ module Hledger.Cli.Commands.Balance (
   balancemode
  ,balance
  ,balanceReportAsText
+ ,balanceReportAsCsv
  ,balanceReportItemAsText
+ ,multiBalanceRowAsCsvText
+ ,multiBalanceRowAsTableText
  ,multiBalanceReportAsText
  ,multiBalanceReportAsCsv
  ,multiBalanceReportAsHtml
  ,multiBalanceReportHtmlRows
+ ,multiBalanceReportHtmlFootRow
  ,balanceReportAsTable
  ,balanceReportTableAsText
  ,tests_Balance
 ) where
 
 import Data.Default (def)
-import Data.List (intersperse, transpose)
-import Data.Maybe (fromMaybe, maybeToList)
+import Data.List (transpose, transpose)
+import qualified Data.Set as S
+import Data.Maybe (fromMaybe)
 import qualified Data.Text as T
 import qualified Data.Text.Lazy as TL
 import qualified Data.Text.Lazy.Builder as TB
-import Data.Time (fromGregorian)
+import Data.Time (addDays, fromGregorian)
 import System.Console.CmdArgs.Explicit as C
-import Lucid as L
-import Text.Tabular.AsciiWide as Tab
+import Lucid as L hiding (value_)
+import Safe (headMay, maximumMay)
+import Text.Tabular.AsciiWide
+    (Align(..), Cell(..), Table(..), TableOpts(..), cellWidth, concatTables,
+    renderColumns, renderRowB, textCell)
+import qualified Text.Tabular.AsciiWide as Tab
 
 import Hledger
 import Hledger.Cli.CliOptions
 import Hledger.Cli.Utils
-import Hledger.Read.CsvReader (CSV, printCSV)
+import Hledger.Read.CsvUtils (CSV, printCSV, printTSV)
 
 
 -- | Command line options for this command.
 balancemode = hledgerCommandMode
   $(embedFileRelative "Hledger/Cli/Commands/Balance.txt")
-  ([flagNone ["sum"] (setboolopt "sum")
+  (
+    -- https://hledger.org/dev/hledger.html#calculation-type :
+    [flagNone ["sum"] (setboolopt "sum")
       "show sum of posting amounts (default)"
-   ,flagNone ["budget"] (setboolopt "budget")
-      "show sum of posting amounts compared to budget goals defined by periodic transactions\n "
-   ,flagNone ["valuechange"] (setboolopt "valuechange")
-      "show change of value of period-end historical balances"
-
-   ,flagNone ["change"] (setboolopt "change")
+    -- XXX --budget[=DESCPAT], --forecast[=PERIODEXP], could be more consistent
+    ,flagOpt "" ["budget"] (\s opts -> Right $ setopt "budget" s opts) "DESCPAT"
+      (unlines
+      [ "show sum of posting amounts together with budget goals defined by periodic"
+      , "transactions. With a DESCPAT argument (must be separated by = not space),"
+      , "use only periodic transactions with matching description"
+      , "(case insensitive substring match)."
+      ])
+    ,flagNone ["valuechange"] (setboolopt "valuechange")
+      "show total change of value of period-end historical balances (caused by deposits, withdrawals, market price fluctuations)"
+    ,flagNone ["gain"] (setboolopt "gain")
+      "show unrealised capital gain/loss (historical balance value minus cost basis)"
+    ,flagNone ["count"] (setboolopt "count") "show the count of postings"
+    -- https://hledger.org/dev/hledger.html#accumulation-type :
+    ,flagNone ["change"] (setboolopt "change")
       "accumulate amounts from column start to column end (in multicolumn reports, default)"
-   ,flagNone ["cumulative"] (setboolopt "cumulative")
+    ,flagNone ["cumulative"] (setboolopt "cumulative")
       "accumulate amounts from report start (specified by e.g. -b/--begin) to column end"
-   ,flagNone ["historical","H"] (setboolopt "historical")
+    ,flagNone ["historical","H"] (setboolopt "historical")
       "accumulate amounts from journal start to column end (includes postings before report start date)\n "
-   ]
-   ++ flattreeflags True ++
-   [flagReq  ["drop"] (\s opts -> Right $ setopt "drop" s opts) "N" "omit N leading account name parts (in flat mode)"
-   ,flagNone ["average","A"] (setboolopt "average") "show a row average column (in multicolumn reports)"
-   ,flagNone ["row-total","T"] (setboolopt "row-total") "show a row total column (in multicolumn reports)"
-   ,flagNone ["no-total","N"] (setboolopt "no-total") "omit the final total row"
-   ,flagNone ["no-elide"] (setboolopt "no-elide") "don't squash boring parent accounts (in tree mode); don't show only 2 commodities per amount"
-   ,flagReq  ["format"] (\s opts -> Right $ setopt "format" s opts) "FORMATSTR" "use this custom line format (in simple reports)"
-   ,flagNone ["pretty-tables"] (setboolopt "pretty-tables") "use unicode to display prettier tables"
-   ,flagNone ["sort-amount","S"] (setboolopt "sort-amount") "sort by amount instead of account code/name (in flat mode). With multiple columns, sorts by the row total, or by row average if that is displayed."
-   ,flagNone ["percent", "%"] (setboolopt "percent") "express values in percentage of each column's total"
-   ,flagNone ["invert"] (setboolopt "invert") "display all amounts with reversed sign"
-   ,flagNone ["transpose"] (setboolopt "transpose") "transpose rows and columns"
-   ,outputFormatFlag ["txt","html","csv","json"]
-   ,outputFileFlag
-   ]
+    ]
+    -- other options specific to this command:
+    ++ flattreeflags True ++
+    [flagReq  ["drop"] (\s opts -> Right $ setopt "drop" s opts) "N" "omit N leading account name parts (in flat mode)"
+    ,flagNone ["declared"] (setboolopt "declared") "include non-parent declared accounts (best used with -E)"
+    ,flagNone ["average","A"] (setboolopt "average") "show a row average column (in multicolumn reports)"
+    ,flagNone ["related","r"] (setboolopt "related") "show postings' siblings instead"
+    ,flagNone ["row-total","T"] (setboolopt "row-total") "show a row total column (in multicolumn reports)"
+    ,flagNone ["summary-only"] (setboolopt "summary-only") "display only row summaries (e.g. row total, average) (in multicolumn reports)"
+    ,flagNone ["no-total","N"] (setboolopt "no-total") "omit the final total row"
+    ,flagNone ["no-elide"] (setboolopt "no-elide") "don't squash boring parent accounts (in tree mode)"
+    ,flagReq  ["format"] (\s opts -> Right $ setopt "format" s opts) "FORMATSTR" "use this custom line format (in simple reports)"
+    ,flagNone ["sort-amount","S"] (setboolopt "sort-amount") "sort by amount instead of account code/name (in flat mode). With multiple columns, sorts by the row total, or by row average if that is displayed."
+    ,flagNone ["percent", "%"] (setboolopt "percent") "express values in percentage of each column's total"
+    ,flagNone ["invert"] (setboolopt "invert") "display all amounts with reversed sign"
+    ,flagNone ["transpose"] (setboolopt "transpose") "transpose rows and columns"
+    ,flagReq  ["layout"] (\s opts -> Right $ setopt "layout" s opts) "ARG"
+      (unlines
+        ["how to lay out multi-commodity amounts and the overall table:"
+        ,"'wide[,WIDTH]': commodities on one line"
+        ,"'tall'        : commodities on separate lines"
+        ,"'bare'        : commodity symbols in one column"
+        ,"'tidy'        : every attribute in its own column"
+        ])
+    -- output:
+    ,outputFormatFlag ["txt","html","csv","tsv","json"]
+    ,outputFileFlag
+    ]
   )
   [generalflagsgroup1]
-  hiddenflags
+  (hiddenflags ++
+    [ flagNone ["commodity-column"] (setboolopt "commodity-column")
+      "show commodity symbols in a separate column, amounts as bare numbers, one row per commodity"
+    ])
   ([], Just $ argsFlag "[QUERY]")
 
 -- | The balance command, prints a balance report.
 balance :: CliOpts -> Journal -> IO ()
-balance opts@CliOpts{reportspec_=rspec} j = case reporttype_ of
-    BudgetReport -> do  -- single or multi period budget report
-      let reportspan = reportSpan j rspec
-          budgetreport = budgetReport rspec (balancingopts_ $ inputopts_ opts) reportspan j
+balance opts@CliOpts{reportspec_=rspec} j = case balancecalc_ of
+    CalcBudget -> do  -- single or multi period budget report
+      let rspan = fst $ reportSpan j rspec
+          budgetreport = styleAmounts styles $ budgetReport rspec (balancingopts_ $ inputopts_ opts) rspan j
           render = case fmt of
             "txt"  -> budgetReportAsText ropts
             "json" -> (<>"\n") . toJsonText
             "csv"  -> printCSV . budgetReportAsCsv ropts
+            "tsv"  -> printTSV . budgetReportAsCsv ropts
             _      -> error' $ unsupportedOutputFormatError fmt
       writeOutputLazyText opts $ render budgetreport
 
     _ | multiperiod -> do  -- multi period balance report
-        let report = multiBalanceReport rspec j
+        let report = styleAmounts styles $ multiBalanceReport rspec j
             render = case fmt of
               "txt"  -> multiBalanceReportAsText ropts
               "csv"  -> printCSV . multiBalanceReportAsCsv ropts
+              "tsv"  -> printTSV . multiBalanceReportAsCsv ropts
               "html" -> (<>"\n") . L.renderText . multiBalanceReportAsHtml ropts
               "json" -> (<>"\n") . toJsonText
               _      -> const $ error' $ unsupportedOutputFormatError fmt  -- PARTIAL:
         writeOutputLazyText opts $ render report
 
     _ -> do  -- single period simple balance report
-        let report = balanceReport rspec j -- simple Ledger-style balance report
+        let report = styleAmounts styles $ balanceReport rspec j -- simple Ledger-style balance report
             render = case fmt of
-              "txt"  -> \ropts -> TB.toLazyText . balanceReportAsText ropts
-              "csv"  -> \ropts -> printCSV . balanceReportAsCsv ropts
+              "txt"  -> \ropts1 -> TB.toLazyText . balanceReportAsText ropts1
+              "csv"  -> \ropts1 -> printCSV . balanceReportAsCsv ropts1
+              "tsv"  -> \ropts1 -> printTSV . balanceReportAsCsv ropts1
               -- "html" -> \ropts -> (<>"\n") . L.renderText . multiBalanceReportAsHtml ropts . balanceReportAsMultiBalanceReport ropts
               "json" -> const $ (<>"\n") . toJsonText
               _      -> error' $ unsupportedOutputFormatError fmt  -- PARTIAL:
         writeOutputLazyText opts $ render ropts report
   where
-    ropts@ReportOpts{..} = rsOpts rspec
-    multiperiod = interval_ /= NoInterval
+    styles = journalCommodityStylesWith HardRounding j
+    ropts@ReportOpts{..} = _rsReportOpts rspec
+    -- Tidy csv/tsv should be consistent between single period and multiperiod reports.
+    multiperiod = interval_ /= NoInterval || (layout_ == LayoutTidy && delimited)
+    delimited   = fmt == "csv" || fmt == "tsv"
     fmt         = outputFormatFromOpts opts
 
 -- XXX this allows rough HTML rendering of a flat BalanceReport, but it can't handle tree mode etc.
@@ -379,31 +420,63 @@ balance opts@CliOpts{reportspec_=rspec} j = case reporttype_ of
 -- | Render a single-column balance report as CSV.
 balanceReportAsCsv :: ReportOpts -> BalanceReport -> CSV
 balanceReportAsCsv opts (items, total) =
-  ["account","balance"] :
-  [[accountNameDrop (drop_ opts) a, wbToText $ showMixedAmountB (balanceOpts False opts) b] | (a, _, _, b) <- items]
-  ++
-  if no_total_ opts
-  then []
-  else [["total", wbToText $ showMixedAmountB (balanceOpts False opts) total]]
+    headers : concatMap (\(a, _, _, b) -> rows a b) items ++ if no_total_ opts then [] else rows "total" total
+  where
+    headers = "account" : case layout_ opts of
+      LayoutBare -> ["commodity", "balance"]
+      _          -> ["balance"]
+    rows :: AccountName -> MixedAmount -> [[T.Text]]
+    rows name ma = case layout_ opts of
+      LayoutBare ->
+          map (\a -> [showName name, acommodity a, renderAmount $ mixedAmount a])
+          . amounts $ mixedAmountStripPrices ma
+      _ -> [[showName name, renderAmount ma]]
+
+    showName = accountNameDrop (drop_ opts)
+    renderAmount amt = wbToText $ showMixedAmountB bopts amt
+      where bopts = csvDisplay{displayOrder = order}
+            order = if layout_ opts == LayoutBare then Just (S.toList $ maCommodities amt) else Nothing
 
 -- | Render a single-column balance report as plain text.
 balanceReportAsText :: ReportOpts -> BalanceReport -> TB.Builder
-balanceReportAsText opts ((items, total)) =
-    unlinesB lines
-    <> unlinesB (if no_total_ opts then [] else [overline, totalLines])
+balanceReportAsText opts ((items, total)) = case layout_ opts of
+    LayoutBare | iscustom -> error' "Custom format not supported with commodity columns"  -- PARTIAL:
+    LayoutBare -> balanceReportAsText' opts ((items, total))
+    _ -> unlinesB ls <> unlinesB (if no_total_ opts then [] else [overline, totalLines])
   where
-    (lines, sizes) = unzip $ map (balanceReportItemAsText opts) items
+    (ls, sizes) = unzip $ map (balanceReportItemAsText opts) items
     -- abuse renderBalanceReportItem to render the total with similar format
     (totalLines, _) = renderBalanceReportItem opts ("",0,total)
     -- with a custom format, extend the line to the full report width;
     -- otherwise show the usual 20-char line for compatibility
-    overlinewidth = case format_ opts of
-        OneLine       ((FormatField _ _ _ TotalField):_) -> 20
-        TopAligned    ((FormatField _ _ _ TotalField):_) -> 20
-        BottomAligned ((FormatField _ _ _ TotalField):_) -> 20
-        _ -> sum (map maximum' $ transpose sizes)
+    iscustom = case format_ opts of
+        OneLine       ((FormatField _ _ _ TotalField):_) -> False
+        TopAligned    ((FormatField _ _ _ TotalField):_) -> False
+        BottomAligned ((FormatField _ _ _ TotalField):_) -> False
+        _ -> True
+    overlinewidth = if iscustom then sum (map maximum' $ transpose sizes) else 20
     overline   = TB.fromText $ T.replicate overlinewidth "-"
 
+-- | Render a single-column balance report as plain text in commodity-column mode
+balanceReportAsText' :: ReportOpts -> BalanceReport -> TB.Builder
+balanceReportAsText' opts ((items, total)) =
+  unlinesB . fmap (renderColumns def{tableBorders=False} sizes .  Tab.Group Tab.NoLine . fmap Tab.Header) $
+    ls ++ concat [[[overline], totalline] | not (no_total_ opts)]
+  where
+    render (_, acctname, dep, amt) =
+        [ Cell TopRight damts
+        , Cell TopLeft (fmap wbFromText cs)
+        , Cell TopLeft (replicate (length damts - 1) mempty ++ [wbFromText dispname]) ]
+      where dopts = oneLine{displayColour=color_ opts, displayOrder=Just cs}
+            cs    = if mixedAmountLooksZero amt then [""] else S.toList $ maCommodities amt
+            dispname = T.replicate ((dep - 1) * 2) " " <> acctname
+            damts = showMixedAmountLinesB dopts amt
+    ls = fmap render items
+    totalline = render ("", "", 0, total)
+    sizes = fromMaybe 0 . maximumMay . map cellWidth <$>
+        transpose ([totalline | not (no_total_ opts)] ++ ls)
+    overline = Cell TopLeft . pure . wbFromText . flip T.replicate "-" . fromMaybe 0 $ headMay sizes
+
 {-
 :r
 This implementation turned out to be a bit convoluted but implements the following algorithm for formatting:
@@ -420,41 +493,38 @@ This implementation turned out to be a bit convoluted but implements the followi
 -- differently-priced quantities of the same commodity will appear merged.
 -- The output will be one or more lines depending on the format and number of commodities.
 balanceReportItemAsText :: ReportOpts -> BalanceReportItem -> (TB.Builder, [Int])
-balanceReportItemAsText opts (_, accountName, depth, amt) =
-  renderBalanceReportItem opts (accountName, depth, amt)
+balanceReportItemAsText opts (_, accountName, dep, amt) =
+  renderBalanceReportItem opts (accountName, dep, amt)
 
 -- | Render a balance report item using the given StringFormat, generating one or more lines of text.
 renderBalanceReportItem :: ReportOpts -> (AccountName, Int, MixedAmount) -> (TB.Builder, [Int])
-renderBalanceReportItem opts (acctname, depth, total) =
+renderBalanceReportItem opts (acctname, dep, total) =
   case format_ opts of
       OneLine       comps -> renderRow' $ render True  True  comps
       TopAligned    comps -> renderRow' $ render True  False comps
       BottomAligned comps -> renderRow' $ render False False comps
   where
     renderRow' is = ( renderRowB def{tableBorders=False, borderSpaces=False}
-                      . Tab.Group NoLine $ map Header is
+                      . Tab.Group Tab.NoLine $ map Tab.Header is
                     , map cellWidth is )
 
-    render topaligned oneline = map (maybeConcat . renderComponent topaligned opts (acctname, depth, total))
-      where maybeConcat (Cell a xs) =
-                if oneline then Cell a [WideBuilder (mconcat . intersperse (TB.fromText ", ") $ map wbBuilder xs) width]
-                           else Cell a xs
-              where width = sumStrict (map ((+2) . wbWidth) xs) -2
-
+    render topaligned oneline = map (renderComponent topaligned oneline opts (acctname, dep, total))
 
 -- | Render one StringFormat component for a balance report item.
-renderComponent :: Bool -> ReportOpts -> (AccountName, Int, MixedAmount) -> StringFormatComponent -> Cell
-renderComponent _ _ _ (FormatLiteral s) = textCell TopLeft s
-renderComponent topaligned opts (acctname, depth, total) (FormatField ljust mmin mmax field) = case field of
+renderComponent :: Bool -> Bool -> ReportOpts -> (AccountName, Int, MixedAmount) -> StringFormatComponent -> Cell
+renderComponent _ _ _ _ (FormatLiteral s) = textCell TopLeft s
+renderComponent topaligned oneline opts (acctname, dep, total) (FormatField ljust mmin mmax field) = case field of
     DepthSpacerField -> Cell align [WideBuilder (TB.fromText $ T.replicate d " ") d]
-                        where d = maybe id min mmax $ depth * fromMaybe 1 mmin
+                        where d = maybe id min mmax $ dep * fromMaybe 1 mmin
     AccountField     -> textCell align $ formatText ljust mmin mmax acctname
     TotalField       -> Cell align . pure $ showMixedAmountB dopts total
     _                -> Cell align [mempty]
   where
-    align = if topaligned then (if ljust then TopLeft    else TopRight)
-                          else (if ljust then BottomLeft else BottomRight)
-    dopts = (balanceOpts True opts){displayOneLine=False, displayMinWidth=mmin, displayMaxWidth=mmax}
+    align | topaligned && ljust = TopLeft
+          | topaligned          = TopRight
+          | ljust               = BottomLeft
+          | otherwise           = BottomRight
+    dopts = noPrice{displayColour=color_ opts, displayOneLine=oneline, displayMinWidth=mmin, displayMaxWidth=mmax}
 
 -- rendering multi-column balance reports
 
@@ -462,69 +532,73 @@ renderComponent topaligned opts (acctname, depth, total) (FormatField ljust mmin
 -- The CSV will always include the initial headings row,
 -- and will include the final totals row unless --no-total is set.
 multiBalanceReportAsCsv :: ReportOpts -> MultiBalanceReport -> CSV
-multiBalanceReportAsCsv opts@ReportOpts{average_, row_total_}
-    (PeriodicReport colspans items (PeriodicReportRow _ coltotals tot avg)) =
-  maybetranspose $
-  ("account" : map showDateSpan colspans
-   ++ ["total"   | row_total_]
-   ++ ["average" | average_]
-  ) :
-  [accountNameDrop (drop_ opts) (displayFull a) :
-   map (wbToText . showMixedAmountB (balanceOpts False opts))
-   (amts
-    ++ [rowtot | row_total_]
-    ++ [rowavg | average_])
-  | PeriodicReportRow a amts rowtot rowavg <- items]
-  ++
-  if no_total_ opts
-  then []
-  else ["total" :
-        map (wbToText . showMixedAmountB (balanceOpts False opts)) (
-          coltotals
-          ++ [tot | row_total_]
-          ++ [avg | average_]
-          )]
+multiBalanceReportAsCsv opts@ReportOpts{..} report = maybeTranspose allRows
   where
-    maybetranspose | transpose_ opts = transpose
-                   | otherwise = id
+    allRows = case layout_ of
+      LayoutTidy -> rows  -- tidy csv should not include totals or averages
+      _ -> rows ++ totals
+    (rows, totals) = multiBalanceReportAsCsv' opts report
+    maybeTranspose = if transpose_ then transpose else id
+
+multiBalanceReportAsCsv' :: ReportOpts -> MultiBalanceReport -> (CSV, CSV)
+multiBalanceReportAsCsv' opts@ReportOpts{..} (PeriodicReport colspans items tr) =
+    (headers : concatMap fullRowAsTexts items, totalrows)
+  where
+    headers = "account" : case layout_ of
+      LayoutTidy -> ["period", "start_date", "end_date", "commodity", "value"]
+      LayoutBare -> "commodity" : dateHeaders
+      _          -> dateHeaders
+    dateHeaders = map showDateSpan colspans ++ ["total" | row_total_] ++ ["average" | average_]
+
+    fullRowAsTexts row = map (showName row :) $ multiBalanceRowAsCsvText opts colspans row
+    showName = accountNameDrop drop_ . prrFullName
+    totalrows
+      | no_total_ = mempty
+      | otherwise = map ("total" :) $ multiBalanceRowAsCsvText opts colspans tr
 
 -- | Render a multi-column balance report as HTML.
 multiBalanceReportAsHtml :: ReportOpts -> MultiBalanceReport -> Html ()
 multiBalanceReportAsHtml ropts mbr =
   let
-    (headingsrow,bodyrows,mtotalsrow) = multiBalanceReportHtmlRows ropts mbr
+    (headingsrow,bodyrows,mtotalsrows) = multiBalanceReportHtmlRows ropts mbr
   in
     table_ $ mconcat $
          [headingsrow]
       ++ bodyrows
-      ++ maybeToList mtotalsrow
+      ++ mtotalsrows
 
 -- | Render the HTML table rows for a MultiBalanceReport.
 -- Returns the heading row, 0 or more body rows, and the totals row if enabled.
-multiBalanceReportHtmlRows :: ReportOpts -> MultiBalanceReport -> (Html (), [Html ()], Maybe (Html ()))
+multiBalanceReportHtmlRows :: ReportOpts -> MultiBalanceReport -> (Html (), [Html ()], [Html ()])
 multiBalanceReportHtmlRows ropts mbr =
   let
-    headingsrow:rest | transpose_ ropts = error' "Sorry, --transpose with HTML output is not yet supported"  -- PARTIAL:
-                     | otherwise = multiBalanceReportAsCsv ropts mbr
-    (bodyrows, mtotalsrow) | no_total_ ropts = (rest,      Nothing)
-                           | otherwise       = (init rest, Just $ last rest)
+    -- TODO: should the commodity_column be displayed as a subaccount in this case as well?
+    (headingsrow:bodyrows, mtotalsrows)
+      | transpose_ ropts = error' "Sorry, --transpose with HTML output is not yet supported"  -- PARTIAL:
+      | otherwise = multiBalanceReportAsCsv' ropts mbr
   in
     (multiBalanceReportHtmlHeadRow ropts headingsrow
     ,map (multiBalanceReportHtmlBodyRow ropts) bodyrows
-    ,multiBalanceReportHtmlFootRow ropts <$> mtotalsrow -- TODO pad totals row with zeros when there are
+    ,multiBalanceReportHtmlFootRow ropts <$> mtotalsrows -- TODO pad totals row with zeros when there are
     )
 
 -- | Render one MultiBalanceReport heading row as a HTML table row.
 multiBalanceReportHtmlHeadRow :: ReportOpts -> [T.Text] -> Html ()
 multiBalanceReportHtmlHeadRow _ [] = mempty  -- shouldn't happen
-multiBalanceReportHtmlHeadRow ropts (acct:rest) =
+multiBalanceReportHtmlHeadRow ropts (acct:cells) =
   let
     defstyle = style_ ""
     (amts,tot,avg)
-      | row_total_ ropts && average_ ropts = (init $ init rest, [last $ init rest], [last rest])
-      | row_total_ ropts                   = (init rest,        [last rest],        [])
-      |                     average_ ropts = (init rest,        [],                 [last rest])
-      | otherwise                          = (rest,             [],                 [])
+      | row_total_ ropts && average_ ropts = (ini2,  sndlst2, lst2)
+      | row_total_ ropts                   = (ini1,  lst1,    [])
+      |                     average_ ropts = (ini1,  [],      lst1)
+      | otherwise                          = (cells, [],      [])
+      where
+        n = length cells
+        (ini1,lst1)    = splitAt (n-1) cells
+        (ini2, rest)   = splitAt (n-2) cells
+        (sndlst2,lst2) = splitAt 1 rest
+
   in
     tr_ $ mconcat $
           td_ [class_ "account"]              (toHtml acct)
@@ -535,14 +609,19 @@ multiBalanceReportHtmlHeadRow ropts (acct:rest) =
 -- | Render one MultiBalanceReport data row as a HTML table row.
 multiBalanceReportHtmlBodyRow :: ReportOpts -> [T.Text] -> Html ()
 multiBalanceReportHtmlBodyRow _ [] = mempty  -- shouldn't happen
-multiBalanceReportHtmlBodyRow ropts (label:rest) =
+multiBalanceReportHtmlBodyRow ropts (label:cells) =
   let
     defstyle = style_ "text-align:right"
     (amts,tot,avg)
-      | row_total_ ropts && average_ ropts = (init $ init rest, [last $ init rest], [last rest])
-      | row_total_ ropts                   = (init rest,        [last rest],        [])
-      |                     average_ ropts = (init rest,        [],                 [last rest])
-      | otherwise                          = (rest,             [],                 [])
+      | row_total_ ropts && average_ ropts = (ini2,  sndlst2, lst2)
+      | row_total_ ropts                   = (ini1,  lst1,    [])
+      |                     average_ ropts = (ini1,  [],      lst1)
+      | otherwise                          = (cells, [],      [])
+      where
+        n = length cells
+        (ini1,lst1)    = splitAt (n-1) cells
+        (ini2, rest)   = splitAt (n-2) cells
+        (sndlst2,lst2) = splitAt 1 rest
   in
     tr_ $ mconcat $
           td_ [class_ "account", style_ "text-align:left"]  (toHtml label)
@@ -559,14 +638,19 @@ multiBalanceReportHtmlFootRow _ropts [] = mempty
 --   : repeat nullmixedamt zeros
 --  ++ (if row_total_ ropts then [nullmixedamt] else [])
 --  ++ (if average_ ropts   then [nullmixedamt]   else [])
-multiBalanceReportHtmlFootRow ropts (acct:rest) =
+multiBalanceReportHtmlFootRow ropts (acct:cells) =
   let
     defstyle = style_ "text-align:right"
     (amts,tot,avg)
-      | row_total_ ropts && average_ ropts = (init $ init rest, [last $ init rest], [last rest])
-      | row_total_ ropts                   = (init rest,        [last rest],        [])
-      |                     average_ ropts = (init rest,        [],                 [last rest])
-      | otherwise                          = (rest,             [],                 [])
+      | row_total_ ropts && average_ ropts = (ini2,  sndlst2, lst2)
+      | row_total_ ropts                   = (ini1,  lst1,    [])
+      |                     average_ ropts = (ini1,  [],      lst1)
+      | otherwise                          = (cells, [],      [])
+      where
+        n = length cells
+        (ini1,lst1)    = splitAt (n-1) cells
+        (ini2, rest)   = splitAt (n-2) cells
+        (sndlst2,lst2) = splitAt 1 rest
   in
     tr_ $ mconcat $
           th_ [style_ "text-align:left"]             (toHtml acct)
@@ -586,16 +670,19 @@ multiBalanceReportAsText ropts@ReportOpts{..} r = TB.toLazyText $
   where
     title = mtitle <> " in " <> showDateSpan (periodicReportSpan r) <> valuationdesc <> ":"
 
-    mtitle = case (reporttype_, balancetype_) of
-        (ValueChangeReport, PeriodChange     ) -> "Period-end value changes"
-        (ValueChangeReport, CumulativeChange ) -> "Cumulative period-end value changes"
-        (_,                 PeriodChange     ) -> "Balance changes"
-        (_,                 CumulativeChange ) -> "Ending balances (cumulative)"
-        (_,                 HistoricalBalance) -> "Ending balances (historical)"
+    mtitle = case (balancecalc_, balanceaccum_) of
+        (CalcValueChange, PerPeriod  ) -> "Period-end value changes"
+        (CalcValueChange, Cumulative ) -> "Cumulative period-end value changes"
+        (CalcGain,        PerPeriod  ) -> "Incremental gain"
+        (CalcGain,        Cumulative ) -> "Cumulative gain"
+        (CalcGain,        Historical ) -> "Historical gain"
+        (_,               PerPeriod  ) -> "Balance changes"
+        (_,               Cumulative ) -> "Ending balances (cumulative)"
+        (_,               Historical)  -> "Ending balances (historical)"
     valuationdesc =
-        (case cost_ of
-            Cost   -> ", converted to cost"
-            NoCost -> "")
+        (case conversionop_ of
+            Just ToCost -> ", converted to cost"
+            _           -> "")
         <> (case value_ of
             Just (AtThen _mc)    -> ", valued at posting date"
             Just (AtEnd _mc) | changingValuation -> ""
@@ -604,65 +691,97 @@ multiBalanceReportAsText ropts@ReportOpts{..} r = TB.toLazyText $
             Just (AtDate d _mc)  -> ", valued at " <> showDate d
             Nothing              -> "")
 
-    changingValuation = case (reporttype_, balancetype_) of
-        (ValueChangeReport, PeriodChange)     -> True
-        (ValueChangeReport, CumulativeChange) -> True
+    changingValuation = case (balancecalc_, balanceaccum_) of
+        (CalcValueChange, PerPeriod)  -> True
+        (CalcValueChange, Cumulative) -> True
         _                                     -> False
 
 -- | Build a 'Table' from a multi-column balance report.
-balanceReportAsTable :: ReportOpts -> MultiBalanceReport -> Table T.Text T.Text MixedAmount
-balanceReportAsTable opts@ReportOpts{average_, row_total_, balancetype_}
-    (PeriodicReport spans items (PeriodicReportRow _ coltotals tot avg)) =
+balanceReportAsTable :: ReportOpts -> MultiBalanceReport -> Table T.Text T.Text WideBuilder
+balanceReportAsTable opts@ReportOpts{summary_only_, average_, row_total_, balanceaccum_}
+    (PeriodicReport spans items tr) =
    maybetranspose $
    addtotalrow $
    Table
-     (Tab.Group NoLine $ map Header accts)
-     (Tab.Group NoLine $ map Header colheadings)
-     (map rowvals items)
+     (Tab.Group Tab.NoLine $ map Tab.Header (concat accts))
+     (Tab.Group Tab.NoLine $ map Tab.Header colheadings)
+     (concat rows)
   where
-    totalscolumn = row_total_ && balancetype_ `notElem` [CumulativeChange, HistoricalBalance]
-    colheadings = map (reportPeriodName balancetype_ spans) spans
+    totalscolumn = row_total_ && balanceaccum_ `notElem` [Cumulative, Historical]
+    colheadings = ["Commodity" | layout_ opts == LayoutBare]
+                  ++ (if not summary_only_ then map (reportPeriodName balanceaccum_ spans) spans else [])
                   ++ ["  Total" | totalscolumn]
                   ++ ["Average" | average_]
-    accts = map renderacct items
+    fullRowAsTexts row =
+      let rs = multiBalanceRowAsTableText opts row
+       in (replicate (length rs) (renderacct row), rs)
+    (accts, rows) = unzip $ fmap fullRowAsTexts items
     renderacct row =
         T.replicate ((prrDepth row - 1) * 2) " " <> prrDisplayName row
-    rowvals (PeriodicReportRow _ as rowtot rowavg) = as
-                             ++ [rowtot | totalscolumn]
-                             ++ [rowavg | average_]
-    addtotalrow | no_total_ opts = id
-                | otherwise      = (+----+ (row "" $
-                                    coltotals
-                                    ++ [tot | totalscolumn && not (null coltotals)]
-                                    ++ [avg | average_   && not (null coltotals)]
-                                    ))
+    addtotalrow
+      | no_total_ opts = id
+      | otherwise =
+        let totalrows = multiBalanceRowAsTableText opts tr
+            rh = Tab.Group Tab.NoLine . replicate (length totalrows) $ Tab.Header ""
+            ch = Tab.Header [] -- ignored
+         in (flip (concatTables Tab.SingleLine) $ Table rh ch totalrows)
     maybetranspose | transpose_ opts = \(Table rh ch vals) -> Table ch rh (transpose vals)
                    | otherwise       = id
 
--- | Given a table representing a multi-column balance report (for example,
--- made using 'balanceReportAsTable'), render it in a format suitable for
--- console output. Amounts with more than two commodities will be elided
--- unless --no-elide is used.
-balanceReportTableAsText :: ReportOpts -> Table T.Text T.Text MixedAmount -> TB.Builder
-balanceReportTableAsText ropts@ReportOpts{..} =
-    Tab.renderTableB def{tableBorders=False, prettyTable=pretty_tables_}
-        (Tab.textCell TopLeft) (Tab.textCell TopRight) $
-        Cell TopRight . pure . showMixedAmountB (balanceOpts True ropts)
+multiBalanceRowAsWbs :: AmountDisplayOpts -> ReportOpts -> [DateSpan] -> PeriodicReportRow a MixedAmount -> [[WideBuilder]]
+multiBalanceRowAsWbs bopts ReportOpts{..} colspans (PeriodicReportRow _ as rowtot rowavg) =
+    case layout_ of
+      LayoutWide width -> [fmap (showMixedAmountB bopts{displayMaxWidth=width}) allamts]
+      LayoutTall       -> paddedTranspose mempty
+                           . fmap (showMixedAmountLinesB bopts{displayMaxWidth=Nothing})
+                           $ allamts
+      LayoutBare       -> zipWith (:) (fmap wbFromText cs)  -- add symbols
+                           . transpose                         -- each row becomes a list of Text quantities
+                           . fmap (showMixedAmountLinesB bopts{displayOrder=Just cs, displayMinWidth=Nothing})
+                           $ allamts
+      LayoutTidy       -> concat
+                           . zipWith (map . addDateColumns) colspans
+                           . fmap ( zipWith (\c a -> [wbFromText c, a]) cs
+                                  . showMixedAmountLinesB bopts{displayOrder=Just cs, displayMinWidth=Nothing})
+                           $ as  -- Do not include totals column or average for tidy output, as this
+                                 -- complicates the data representation and can be easily calculated
+  where
+    totalscolumn = row_total_ && balanceaccum_ `notElem` [Cumulative, Historical]
+    cs = if all mixedAmountLooksZero allamts then [""] else S.toList $ foldMap maCommodities allamts
+    allamts = (if not summary_only_ then as else []) ++
+                [rowtot | totalscolumn && not (null as)] ++
+                [rowavg | average_ && not (null as)]
+    addDateColumns spn@(DateSpan s e) = (wbFromText (showDateSpan spn) :)
+                                       . (wbFromText (maybe "" showEFDate s) :)
+                                       . (wbFromText (maybe "" (showEFDate . modifyEFDay (addDays (-1))) e) :)
 
--- | Amount display options to use for balance reports
-balanceOpts :: Bool -> ReportOpts -> AmountDisplayOpts
-balanceOpts isTable ReportOpts{..} = oneLine
-    { displayColour   = isTable && color_
-    , displayMaxWidth = if isTable && not no_elide_ then Just 32 else Nothing
-    }
+    paddedTranspose :: a -> [[a]] -> [[a]]
+    paddedTranspose _ [] = [[]]
+    paddedTranspose n as1 = take (maximum . map length $ as1) . trans $ as1
+        where
+          trans ([] : xss)  = (n : map h xss) :  trans ([n] : map t xss)
+          trans ((x : xs) : xss) = (x : map h xss) : trans (m xs : map t xss)
+          trans [] = []
+          h (x:_) = x
+          h [] = n
+          t (_:xs) = xs
+          t [] = [n]
+          m (x:xs) = x:xs
+          m [] = [n]
 
-tests_Balance = tests "Balance" [
+multiBalanceRowAsCsvText :: ReportOpts -> [DateSpan] -> PeriodicReportRow a MixedAmount -> [[T.Text]]
+multiBalanceRowAsCsvText opts colspans = fmap (fmap wbToText) . multiBalanceRowAsWbs csvDisplay opts colspans
 
-   tests "balanceReportAsText" [
-    test "unicode in balance layout" $ do
+multiBalanceRowAsTableText :: ReportOpts -> PeriodicReportRow a MixedAmount -> [[WideBuilder]]
+multiBalanceRowAsTableText opts = multiBalanceRowAsWbs oneLine{displayColour=color_ opts} opts []
+
+tests_Balance = testGroup "Balance" [
+
+   testGroup "balanceReportAsText" [
+    testCase "unicode in balance layout" $ do
       j <- readJournal' "2009/01/01 * медвежья шкура\n  расходы:покупки  100\n  актив:наличные\n"
-      let rspec = defreportspec{rsOpts=defreportopts{no_total_=True}}
-      TB.toLazyText (balanceReportAsText (rsOpts rspec) (balanceReport rspec{rsToday=fromGregorian 2008 11 26} j))
+      let rspec = defreportspec{_rsReportOpts=defreportopts{no_total_=True}}
+      TB.toLazyText (balanceReportAsText (_rsReportOpts rspec) (balanceReport rspec{_rsDay=fromGregorian 2008 11 26} j))
         @?=
         TL.unlines
         ["                -100  актив:наличные"
